/*
    DPS - Dynamic Parallel Schedules
    Copyright (C) 2000-2003 Sebastian Gerlach

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
	DPS Layer 1 - Base library

	This file is part of the base namespace, containing all the basic 
	functionality used by DPS. It provides support for reflection, smart
	pointers, serialization, and containers. It also contains cross-platform
	abstractions for several useful system services, such as threads, sockets,
	and synchronization primitives.
	The classes in this namespace have no dependencies on other DPS classes.
*/

/*! \file basesync.h
	\brief Synchronization and threading classes

	Contains classes for abstracting threads and synchronization primitives.
	This is where most of the system dependent code of DPS resides. The model
	is based essentially on the Win32 threading and synchronization API, with
	a similar functionality implemented on top of pthreads for UNIX platforms.
*/

#ifndef INCLUDED_DPS_SYNCHRO_H
#define INCLUDED_DPS_SYNCHRO_H

#define WAIT_INFINITE		0xffffffff
#define WAIT_ERROR			0xffffffff

#define THREAD_TIMEOUT		10000
#define MAX_THREAD_DELAY	300

namespace base
{
	//! Type for thread callback function
	typedef Int32 (*ThreadFunc)(void *arg);
}	

#ifdef DPS_PLATFORM_WIN32
namespace base
{
	// Win32 version

	//! Wrapper class for critical sections
	/*! \ingroup internal
	*/
	class CriticalSection
	{
		//! Win32 critical section object
		CRITICAL_SECTION cs;

	public:
		//! Constructor
		CriticalSection() { InitializeCriticalSection(&cs); }
		//! Destructor
		~CriticalSection() { DeleteCriticalSection(&cs); }
		
		//! Enter critical section
		void enter() { EnterCriticalSection(&cs); }
		//! Leave critical section
		void leave() { LeaveCriticalSection(&cs); }
	};

	//! Wrapper class for events (in the Win32 sense)
	/*! \ingroup internal
	*/
	class Event
	{
		//! Event handle
		HANDLE hEvent;
		
	public:
		//! Constructor
		/*! \param manualReset true if this is a manual reset event
		*/
		Event(Bool manualReset=false) 
		{ 
			hEvent=CreateEvent(NULL,manualReset,FALSE,NULL); 
		}

		//! Destructor
		~Event() 
		{ 
			CloseHandle(hEvent); 
		}

		//! Wait for event
		/*! \param timeout Timeout in milliseconds
			\return WAIT_OBJECT_0 if event is signaled, WAIT_TIMEOUT if timeout is reached, WAIT_ERROR if there was an error
		*/
		UInt32 wait(UInt32 timeout) 
		{ 
			// Return WAIT_OBJECT_0 or WAIT_TIMEOUT
			return WaitForSingleObject(hEvent,timeout==WAIT_INFINITE ? INFINITE : timeout); 
		}

		//! Set the event
		UInt32 set() 
		{ 
			return SetEvent(hEvent); 
		}

		//! Reset the event
		UInt32 reset() 
		{ 
			return ResetEvent(hEvent); 
		}
	};

	//! Thread wrapper class
	/*! \ingroup internal
	*/
	class SysThread
	{
	private:
		//! Thread handle
		HANDLE hThread;
		//! Thread identifier
		DWORD tid;
		//! Root fiber for the thread
		LPVOID rootFiber;
		//! Pointer to thread function
		ThreadFunc threadFunction;
		//! Argument to thread function
		void *threadArgument;
		
	protected:
		//! Event to indicate that thread termination is desired
		Event terminateEvent;
		
	public:
		//! Constructor
		/*! \param threadFunc Pointer to thread function
			\param argument Thread function argument
		*/
		SysThread(ThreadFunc threadFunc, void *argument) : terminateEvent(true) 
		{ 
			threadFunction=threadFunc;
			threadArgument=argument;
			hThread=NULL; 
			rootFiber=NULL; 
		}
		//! Destructor
		~SysThread() 
		{ 
			if(hThread) 
				stop(); 
		}
		//! Thread wrapper function
		static DWORD WINAPI threadWrapper(LPVOID arg)
		{
			SysThread *t=(SysThread*)arg;
			t->rootFiber=ConvertThreadToFiber(NULL);
			return (DWORD)t->threadFunction(t->threadArgument);
		}

		//! Start thread
		Result start();

		//! Stop thread
		Result stop(Bool fast=false);

		//! Wait for end of thread
		Result wait();

		//! Should thread stop?
		bool shouldStop() { return terminateEvent.wait(0)==WAIT_OBJECT_0; }

		//! Sleep function
		/*! \param ms Time to sleep in milliseconds
		*/
		static void sleep(Int32 ms) { Sleep(ms); }

		//! Get root fiber for this thread
		void *getRootFiber() { return rootFiber; }

		//! Exit current thread
		static void exit(int rv) { ExitThread((UInt32)rv); }
	};

	//! Exec function wrapper
	/*! \ingroup internal
	*/
	class ExecHelper
	{
	public:
		//! Execute a command
		/*! \param cmd Command line to execute
		*/
		static UInt32 exec(const char *cmd);

		//! Get name of this executable
		static std::string getExecutableName();

		//! Get process id of current process
		static UInt32 getPid() { return GetCurrentProcessId(); }

		//! Kill process
		/*! \param pid Process id of process to kill
		*/
		static void kill(UInt32 pid);
	};

	//! Shared library wrapper
	/*! \ingroup internal
	*/
	class SharedLibrary
	{
		//! Handle to a shared library
		HMODULE hLib;
	public:
		//! Constructor
		/*! \param libName Filename of shared library to load
		*/
		SharedLibrary(const char *libName);

		//! Destructor
		~SharedLibrary();

		//! Get address of a procedure
		/*! \param procName Name of procedure
		*/
		void *getProcAddress(const char *procName);
	};

	//! Timer class 
	/*! \ingroup internal
	*/
	class Timer
	{
		//! Startup time
		static Int64 baseTime;
		//! Timer frequency
		static Int64 frequency;
	public:
		//! Start timer 
		static void start() { QueryPerformanceCounter((LARGE_INTEGER*)&baseTime); QueryPerformanceFrequency((LARGE_INTEGER*)&frequency); } 
		//! Reset startup time
		static void reset(Int64 bt) { Int64 nbt; QueryPerformanceCounter((LARGE_INTEGER*)&nbt); baseTime=nbt-bt; }
		//! Get current timer value
		static Int64 get() { Int64 nbt; QueryPerformanceCounter((LARGE_INTEGER*)&nbt); return nbt-baseTime; }
		//! Get timer frequency
		static Int64 getFrequency() { return frequency; }
	};

}
#else

#define WAIT_OBJECT_0		0x00000001
#define WAIT_TIMEOUT		0x00000000

#include <pthread.h>
#include <sys/timeb.h>
#include <unistd.h>
#include <errno.h>
#include <sys/time.h>
#include <signal.h>

namespace base
{

	//! Wrapper class for critical sections (pthreads version)
	/*! \ingroup internal
		This is slightly longer than the Win32 version since pthread mutexes
		are not recursive by default. Some OSes provide extensions for this
		(e.g. Linux), but not all.
	*/
	class CriticalSection
	{
		//! Pthreads mutex
		pthread_mutex_t cs;
		//! Owner thread
		pthread_t owner;
		//! Recursive count
		volatile Int32 recursive;
		
	public:
		//! Constructor
		CriticalSection()
		{ 
			owner=0;
			recursive=0;
			pthread_mutex_init(&cs,NULL); 
		}
		//! Destructor
		~CriticalSection() 
		{
			pthread_mutex_destroy(&cs);
		}
		
		//! Enter critical section
		void enter() 
		{ 
			pthread_t tt=pthread_self();
			if(tt==owner)
				recursive++;
			else if(pthread_mutex_lock(&cs)==0)
			{
				owner=tt;
				recursive=0;
			}
			else
			{
				// TODO: failure condition
			}
		}

		//! Leave critical section
		void leave() 
		{ 
			if(owner==pthread_self())
			{
				if(recursive)
					recursive--;
				else
				{
					owner=0;
					pthread_mutex_unlock(&cs); 
				}
			}
			else
			{
				// TODO: failure condition
			}
		}
	};


	//! Event according to Win32 (pthreads version)
	/*! \ingroup internal
		Simulates a Win32 Event object using the primitives provided by
		pthreads.
	*/
	class Event
	{
		//! Condition variable
		pthread_cond_t theEvent;
		//! Mutex
		pthread_mutex_t theMutex;
		//! Boolean indicating state of event
		volatile Bool isSet;
		//! Boolean indicating manual reset
		Bool manualReset;
		
	public:
		//! Constructor
		Event(Bool mr=false) 
		{ 
			pthread_cond_init(&theEvent,NULL); 
			pthread_mutex_init(&theMutex,NULL); 
			manualReset=mr;
			isSet=false;
		}

		//! Destructor
		~Event() 
		{ 
			pthread_cond_destroy(&theEvent); 
			pthread_mutex_destroy(&theMutex); 
		}

		//! Wait for an event
		Int32 wait(Int32 timeout) 
		{ 
			pthread_mutex_lock(&theMutex);
			if(isSet)
			{
				if(!manualReset)
					isSet=false;
				pthread_mutex_unlock(&theMutex);
				return WAIT_OBJECT_0;
			}
			if(timeout==0)
			{
				pthread_mutex_unlock(&theMutex);
				return WAIT_TIMEOUT;
			}

			if(timeout==(Int32)WAIT_INFINITE) 
			{
				while(!isSet)
				{
					pthread_cond_wait(&theEvent,&theMutex); 
				}
				if(!manualReset)
					isSet=false;
				pthread_mutex_unlock(&theMutex);
				return WAIT_OBJECT_0;
			}

			int retval;
			struct timeval delta;
			struct timespec abstime;
			
			gettimeofday(&delta, NULL);
			
			abstime.tv_sec = delta.tv_sec + (timeout/1000);
			abstime.tv_nsec = (delta.tv_usec+(timeout%1000)*1000) * 1000;
			if ( abstime.tv_nsec > 1000000000 ) 
			{
				abstime.tv_sec += 1;
				abstime.tv_nsec -= 1000000000;
			}
			
tryagain:
			retval = pthread_cond_timedwait(&theEvent, &theMutex, &abstime);
			switch (retval) 
			{
			case EINTR:
				goto tryagain;
				break;
			case ETIMEDOUT:
				retval = WAIT_TIMEOUT;
				break;
			case 0:
				{
					if(isSet)
					{
						retval=WAIT_OBJECT_0;
						if(!manualReset)
							isSet=false;
					}
					else
						retval=WAIT_TIMEOUT;
				}
				break;
			default:				
				retval=WAIT_ERROR;
				break;
			}
			pthread_mutex_unlock(&theMutex);
			return retval;
		}

		//! Set an event
		UInt32 set() 
		{ 
			pthread_mutex_lock(&theMutex);
			isSet=true;
			pthread_cond_signal(&theEvent); 
			pthread_mutex_unlock(&theMutex);
			return 0;
		}

		//! Reset the event
		UInt32 reset() 
		{ 
			pthread_mutex_lock(&theMutex);
			isSet=false;
			pthread_mutex_unlock(&theMutex);
			return 0;
		}
	};
	
	//! Wrapper class for threads (pthreads version)
	/*! \ingroup internal
	*/
	class SysThread
	{
		//! Pthreads thread handle
		pthread_t theThread;

		//! Pointer to thread function
		ThreadFunc threadFunction;

		//! Argument to thread function
		void *threadArgument;
		
	protected:
		//! Event set to indicate that thread should terminate
		Event terminateEvent;

		//! Event set to indicate that the thread has terminated
		Event finalEvent;

	private:
		//! Wrapper for thread function
		static void * threadFuncWrapper(void *arg)
		{
			SysThread *t=(SysThread*)arg;
			void *rv=(void*)t->threadFunction(t->threadArgument);
			t->finalEvent.set();
			return rv;
		}
		
	public:
		//! Constructor
		/*! \param threadFunc Pointer to thread function
			\param argument Thread function argument
		*/
		SysThread(ThreadFunc threadFunc, void *argument) : terminateEvent(true), finalEvent(true) 
		{ 
			threadFunction=threadFunc;
			threadArgument=argument;
			theThread=0; 
		}

		//! Destructor
		virtual ~SysThread() { if(theThread) stop(); }
		
		//! Start thread
		Result start();

		//! Stop thread
		Result stop(Bool fast = false);

		//! Wait until thread has terminated
		Result wait();

		//! Should this thread stop?
		Bool shouldStop() { return terminateEvent.wait(0)==WAIT_OBJECT_0; }

		//! Sleep function
		static void sleep(Int32 ms) { usleep(ms*1000); }

		//! Exit current thread
		static void exit(int rv) { pthread_exit((void*)rv); }
	};


	//! Exec function wrapper
	/*! \ingroup internal
	*/
	class ExecHelper
	{
	public:
		//! Execute a command
		/*! \param cmd Command line to execute
		*/
		static UInt32 exec(const char *cmd);

		//! Get name of this executable
		static std::string getExecutableName();

		//! Get process id of current process
		static UInt32 getPid() { return getpid(); }

		//! Kill process
		/*! \param pid Process id of process to kill
		*/
		static void kill(UInt32 pid);
	};

	//! Timer class 
	/*! \ingroup internal
	*/
	class Timer
	{
		//! Startup time
		static Int64 baseTime;
		//! Timer frequency
		static Int64 frequency;

		//! Internal get time function
		static Int64 innerGet() 
		{
			struct timeval delta;
			gettimeofday(&delta, NULL);
			return (Int64)delta.tv_sec*1000000+delta.tv_usec;
		}

	public:
		//! Start timer 
		static void start() { baseTime=innerGet(); frequency=1000000; }
		//! Reset startup time
		static void reset(Int64 bt) { baseTime=innerGet()-bt; }
		//! Get current timer value
		static Int64 get() { return innerGet()-baseTime; }
		//! Get timer frequency
		static Int64 getFrequency() { return 1000000; }
	};

}
#endif

namespace base
{
	//! Scoped critical section control
	/*! \ingroup internal
	*/
	class ScopedLock
	{
		//! Critical section to use
		CriticalSection& critSec;

	public:
		//! Constructor
		/*! \param cs Critical section to use.
			The provided critical section is entered.
		*/
		ScopedLock(CriticalSection& cs) : critSec(cs)
		{
			critSec.enter();
		}
		//! Destructor
		/*! Leaves the critical section
		*/
		~ScopedLock()
		{
			critSec.leave();
		}
	};

//! Internal helper thread creation macro
/*! \ingroup internal
*/
#define THREAD_DECL(tn,tf) base::SysThread tn; static Int32 tn##Start(void *arg) { return ((containerType*)arg)->tf(); }
//! Internal helper thread constructor macro
/*! \ingroup internal
*/
#define THREAD_CONS(tn) tn(tn##Start, this)

}

#endif

